<?php

namespace Drupal\Tests\cached_moderation_state\Kernel;

use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\FieldDefinitionInterface;

use Drupal\cached_moderation_state\Plugin\Field\FieldType\CachedModerationStateItem;

use Drupal\KernelTests\KernelTestBase;

use Drupal\Tests\cached_moderation_state\Traits\ContentModerationHelperTrait;
use Drupal\Tests\cached_moderation_state\Traits\InstallModulesTrait;

/**
 * Tests the field config handler service and related hooks.
 *
 * Copyright (C) 2022  Library Solutions, LLC (et al.).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * @group cached_moderation_state
 */
class FieldConfigHandlerTest extends KernelTestBase {

  use ContentModerationHelperTrait;
  use InstallModulesTrait;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();

    $this->installModules([
      'content_moderation',
      'node',
    ]);
  }

  /**
   * Test moderated entities have a 'cached_moderation_state' item on creation.
   */
  public function testFieldItemCreationOnEntityCreate() {
    $node_type = $this->createModeratedContentType();
    $bundle = $node_type->id();

    $this->installModules(['cached_moderation_state']);

    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface */
    $entity_type_manager = $this->container->get('entity_type.manager');
    $node_storage = $entity_type_manager->getStorage('node');

    /** @var \Drupal\node\NodeInterface */
    $node = $node_storage->create([
      'title' => $this->randomMachineName(),
      'type' => $bundle,
    ]);

    $this->assertInstanceOf(FieldableEntityInterface::class, $node);
    $this->assertTrue($node->hasField('cached_moderation_state'));

    $this->assertInstanceOf(CachedModerationStateItem::class, $node->cached_moderation_state->first());
  }

  /**
   * Test moderated entities have a 'cached_moderation_state' item on load.
   */
  public function testFieldItemCreationOnEntityLoad() {
    $node_type = $this->createModeratedContentType();
    $bundle = $node_type->id();

    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface */
    $entity_type_manager = $this->container->get('entity_type.manager');
    $node_storage = $entity_type_manager->getStorage('node');

    /** @var \Drupal\node\NodeInterface */
    $node = $node_storage->create([
      'title' => $this->randomMachineName(),
      'type' => $bundle,
    ]);

    $node->save();
    $node = $node->id();
    $node_storage->resetCache();

    $this->installModules(['cached_moderation_state']);

    /** @var \Drupal\node\NodeInterface */
    $node = $node_storage->load($node);

    $this->assertInstanceOf(FieldableEntityInterface::class, $node);
    $this->assertTrue($node->hasField('cached_moderation_state'));

    $this->assertInstanceOf(CachedModerationStateItem::class, $node->cached_moderation_state->first());
  }

  /**
   * Test 'cached_moderation_state' is added to moderated bundles on install.
   */
  public function testFieldSyncOnInstall() {
    $moderated = $this->createModeratedContentType()->id();
    $unmoderated = $this->createContentType()->id();

    $this->installModules(['cached_moderation_state']);

    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface */
    $entity_type_manager = $this->container->get('entity_type.manager');
    $field_config_storage = $entity_type_manager->getStorage('field_config');

    $moderated_field = $field_config_storage->load("node.{$moderated}.cached_moderation_state");
    $unmoderated_field = $field_config_storage->load("node.{$unmoderated}.cached_moderation_state");

    $this->assertInstanceOf(FieldDefinitionInterface::class, $moderated_field);
    $this->assertEmpty($unmoderated_field);
  }

  /**
   * Test 'cached_moderation_state' is added when inserting a workflow.
   */
  public function testFieldSyncOnWorkflowInsert() {
    $node_type = $this->createContentType();
    $bundle = $node_type->id();

    $this->installModules(['cached_moderation_state']);

    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface */
    $entity_type_manager = $this->container->get('entity_type.manager');
    $field_config_storage = $entity_type_manager->getStorage('field_config');

    $field = $field_config_storage->load("node.{$bundle}.cached_moderation_state");
    $this->assertEmpty($field);

    $this->createModerationWorkflow([$bundle]);

    $field = $field_config_storage->load("node.{$bundle}.cached_moderation_state");
    $this->assertInstanceOf(FieldDefinitionInterface::class, $field);
  }

  /**
   * Test 'cached_moderation_state' is added when updating a workflow.
   */
  public function testFieldSyncOnWorkflowUpdate() {
    $workflow = $this->createModerationWorkflow();
    $node_type = $this->createContentType();
    $bundle = $node_type->id();

    $this->installModules(['cached_moderation_state']);

    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface */
    $entity_type_manager = $this->container->get('entity_type.manager');
    $field_config_storage = $entity_type_manager->getStorage('field_config');

    $field = $field_config_storage->load("node.{$bundle}.cached_moderation_state");
    $this->assertEmpty($field);

    /** @var \Drupal\content_moderation\Plugin\WorkflowType\ContentModerationInterface */
    $workflow_type = $workflow->getTypePlugin();

    $workflow_type->addEntityTypeAndBundle('node', $bundle);
    $workflow->save();

    $field = $field_config_storage->load("node.{$bundle}.cached_moderation_state");
    $this->assertInstanceOf(FieldDefinitionInterface::class, $field);

    $workflow_type->removeEntityTypeAndBundle('node', $bundle);
    $workflow->save();

    $field = $field_config_storage->load("node.{$bundle}.cached_moderation_state");
    $this->assertEmpty($field);
  }

}
